@@ -3,6 +3,7 @@
  *
  * Copyright 1997,1998 Pavel Machek, distribute under GPL
  *  <pavel@atrey.karlin.mff.cuni.cz>
+ * Copyright (c) 2002 - 2011 Wouter Verhelst <w@uter.be>
  *
  * Version 1.0 - 64bit issues should be fixed, now
  * Version 1.1 - added bs (blocksize) option (Alexey Guzeev, aga@permonline.ru)
@@ -27,12 +28,14 @@
 #include <netinet/tcp.h>
 #include <netinet/in.h>
 #include <netdb.h>
+#include "netdb-compat.h"
 #include <stdio.h>
 #include <fcntl.h>
 #include <syslog.h>
 #include <stdlib.h>
 #include <sys/mount.h>
 #include <sys/mman.h>
+#include <signal.h>
 #include <errno.h>
 #include <getopt.h>
 #include <stdarg.h>
@@ -45,14 +48,16 @@
 #include <sdp_inet.h>
 #endif
 
+#define NBDC_DO_LIST 1
+
 int check_conn(char* devname, int do_print) {
 	char buf[256];
 	char* p;
 	int fd;
 	int len;
 
-	if(!strncmp(devname, "/dev/", 5)) {
-		devname+=5;
+	if( (p=strrchr(devname, '/')) ) {
+		devname=p+1;
 	}
 	if((p=strchr(devname, 'p'))) {
 		/* We can't do checks on partitions. */
@@ -90,7 +95,7 @@ int opennet(char *name, char* portstr, int sdp) {
 	if(e != 0) {
 		fprintf(stderr, "getaddrinfo failed: %s\n", gai_strerror(e));
 		freeaddrinfo(ai);
-		exit(EXIT_FAILURE);
+		return -1;
 	}
 
 	if(sdp) {
@@ -114,8 +119,10 @@ int opennet(char *name, char* portstr, int sdp) {
 			break;		/* success */
 	}
 
-	if (rp == NULL)
-		err("Socket failed: %m");
+	if (rp == NULL) {
+		err_nonfatal("Socket failed: %m");
+		return -1;
+	}
 
 	setmysockopt(sock);
 
@@ -123,7 +130,102 @@ int opennet(char *name, char* portstr, int sdp) {
 	return sock;
 }
 
-void negotiate(int sock, u64 *rsize64, u32 *flags, char* name) {
+void ask_list(int sock) {
+	uint32_t opt;
+	uint32_t opt_server;
+	uint32_t len;
+	uint32_t reptype;
+	uint64_t magic;
+	const int BUF_SIZE = 1024;
+	char buf[BUF_SIZE];
+
+	magic = ntohll(opts_magic);
+	if (write(sock, &magic, sizeof(magic)) < 0)
+		err("Failed/2.2: %m");
+
+	/* Ask for the list */
+	opt = htonl(NBD_OPT_LIST);
+	if(write(sock, &opt, sizeof(opt)) < 0) {
+		err("writing list option failed: %m");
+	}
+	/* Send the length (zero) */
+	len = htonl(0);
+	if(write(sock, &len, sizeof(len)) < 0) {
+		err("writing length failed: %m");
+	}
+	/* newline, move away from the "Negotiation:" line */
+	printf("\n");
+	do {
+		memset(buf, 0, 1024);
+		if(read(sock, &magic, sizeof(magic)) < 0) {
+			err("Reading magic from server: %m");
+		}
+		if(read(sock, &opt_server, sizeof(opt_server)) < 0) {
+			err("Reading option: %m");
+		}
+		if(read(sock, &reptype, sizeof(reptype)) <0) {
+			err("Reading reply from server: %m");
+		}
+		if(read(sock, &len, sizeof(len)) < 0) {
+			err("Reading length from server: %m");
+		}
+		magic=ntohll(magic);
+		len=ntohl(len);
+		reptype=ntohl(reptype);
+		if(magic != rep_magic) {
+			err("Not enough magic from server");
+		}
+		if(reptype & NBD_REP_FLAG_ERROR) {
+			switch(reptype) {
+				case NBD_REP_ERR_POLICY:
+					fprintf(stderr, "\nE: listing not allowed by server.\n");
+					break;
+				default:
+					fprintf(stderr, "\nE: unexpected error from server.\n");
+					break;
+			}
+			if(len) {
+				if(read(sock, buf, len) < 0) {
+					fprintf(stderr, "\nE: could not read error message from server\n");
+				}
+				fprintf(stderr, "Server said: %s\n", buf);
+			}
+			exit(EXIT_FAILURE);
+		} else {
+			if(len) {
+				if(reptype != NBD_REP_SERVER) {
+					err("Server sent us a reply we don't understand!");
+				}
+				if(read(sock, &len, sizeof(len)) < 0) {
+					fprintf(stderr, "\nE: could not read export name length from server\n");
+					exit(EXIT_FAILURE);
+				}
+				len=ntohl(len);
+				if (len >= BUF_SIZE) {
+					fprintf(stderr, "\nE: export name on server too long\n");
+					exit(EXIT_FAILURE);
+				}
+				if(read(sock, buf, len) < 0) {
+					fprintf(stderr, "\nE: could not read export name from server\n");
+					exit(EXIT_FAILURE);
+				}
+				buf[len] = 0;
+				printf("%s\n", buf);
+			}
+		}
+	} while(reptype != NBD_REP_ACK);
+	opt=htonl(NBD_OPT_ABORT);
+	len=htonl(0);
+	magic=htonll(opts_magic);
+	if (write(sock, &magic, sizeof(magic)) < 0)
+		err("Failed/2.2: %m");
+	if (write(sock, &opt, sizeof(opt)) < 0)
+		err("Failed writing abort");
+	if (write(sock, &len, sizeof(len)) < 0)
+		err("Failed writing length");
+}
+
+void negotiate(int sock, u64 *rsize64, u32 *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts) {
 	u64 magic, size64;
 	uint16_t tmp;
 	char buf[256] = "\0\0\0\0\0\0\0\0\0";
@@ -142,24 +244,39 @@ void negotiate(int sock, u64 *rsize64, u32 *flags, char* name) {
 	if(name) {
 		uint32_t opt;
 		uint32_t namesize;
-		uint32_t reserved = 0;
 
-		if (magic != opts_magic)
-			err("Not enough opts_magic");
+		if (magic != opts_magic) {
+			if(magic == cliserv_magic) {
+				err("It looks like you're trying to connect to an oldstyle server with a named export. This won't work.");
+			}
+		}
 		printf(".");
 		if(read(sock, &tmp, sizeof(uint16_t)) < 0) {
 			err("Failed reading flags: %m");
 		}
 		*flags = ((u32)ntohs(tmp));
+		if((needed_flags & *flags) != needed_flags) {
+			/* There's currently really only one reason why this
+			 * check could possibly fail, but we may need to change
+			 * this error message in the future... */
+			fprintf(stderr, "\nE: Server does not support listing exports\n");
+			exit(EXIT_FAILURE);
+		}
 
-		/* reserved for future use*/
-		if (write(sock, &reserved, sizeof(reserved)) < 0)
+		client_flags = htonl(client_flags);
+		if (write(sock, &client_flags, sizeof(client_flags)) < 0)
 			err("Failed/2.1: %m");
 
+		if(do_opts & NBDC_DO_LIST) {
+			ask_list(sock);
+			exit(EXIT_SUCCESS);
+		}
+
 		/* Write the export name that we're after */
-		magic = ntohll(opts_magic);
+		magic = htonll(opts_magic);
 		if (write(sock, &magic, sizeof(magic)) < 0)
 			err("Failed/2.2: %m");
+
 		opt = ntohl(NBD_OPT_EXPORT_NAME);
 		if (write(sock, &opt, sizeof(opt)) < 0)
 			err("Failed/2.3: %m");
@@ -170,28 +287,27 @@ void negotiate(int sock, u64 *rsize64, u32 *flags, char* name) {
 		if (write(sock, name, strlen(name)) < 0)
 			err("Failed/2.4: %m");
 	} else {
-		if (magic != cliserv_magic)
-			err("Not enough cliserv_magic");
+		if (magic != cliserv_magic) {
+			if(magic != opts_magic)
+				err("Not enough cliserv_magic");
+			else
+				err("It looks like you're trying to connect to a newstyle server with the oldstyle protocol. Try the -N option.");
+		}
 		printf(".");
 	}
 
-	if (read(sock, &size64, sizeof(size64)) < 0)
+	if (read(sock, &size64, sizeof(size64)) <= 0) {
+		if (!errno)
+			err("Server closed connection");
 		err("Failed/3: %m\n");
+	}
 	size64 = ntohll(size64);
 
-#ifdef NBD_SET_SIZE_BLOCKS
 	if ((size64>>12) > (uint64_t)~0UL) {
 		printf("size = %luMB", (unsigned long)(size64>>20));
 		err("Exported device is too big for me. Get 64-bit machine :-(\n");
 	} else
 		printf("size = %luMB", (unsigned long)(size64>>20));
-#else
-	if (size64 > (~0UL >> 1)) {
-		printf("size = %luKB", (unsigned long)(size64>>10));
-		err("Exported device is too big. Get 64-bit machine or newer kernel :-(\n");
-	} else
-		printf("size = %lu", (unsigned long)(size64));
-#endif
 
 	if(!name) {
 		if (read(sock, flags, sizeof(*flags)) < 0)
@@ -214,29 +330,18 @@ void setsizes(int nbd, u64 size64, int blocksize, u32 flags) {
 	unsigned long size;
 	int read_only = (flags & NBD_FLAG_READ_ONLY) ? 1 : 0;
 
-#ifdef NBD_SET_SIZE_BLOCKS
 	if (size64>>12 > (uint64_t)~0UL)
 		err("Device too large.\n");
 	else {
-		int er;
 		if (ioctl(nbd, NBD_SET_BLKSIZE, 4096UL) < 0)
 			err("Ioctl/1.1a failed: %m\n");
 		size = (unsigned long)(size64>>12);
-		if ((er = ioctl(nbd, NBD_SET_SIZE_BLOCKS, size)) < 0)
+		if (ioctl(nbd, NBD_SET_SIZE_BLOCKS, size) < 0)
 			err("Ioctl/1.1b failed: %m\n");
 		if (ioctl(nbd, NBD_SET_BLKSIZE, (unsigned long)blocksize) < 0)
 			err("Ioctl/1.1c failed: %m\n");
 		fprintf(stderr, "bs=%d, sz=%llu bytes\n", blocksize, 4096ULL*size);
 	}
-#else
-	if (size64 > (~0UL >> 1)) {
-		err("Device too large.\n");
-	} else {
-		size = (unsigned long)size64;
-		if (ioctl(nbd, NBD_SET_SIZE, size) < 0)
-			err("Ioctl NBD_SET_SIZE failed: %m\n");
-	}
-#endif
 
 	ioctl(nbd, NBD_CLEAR_SOCK);
 
@@ -249,13 +354,9 @@ void setsizes(int nbd, u64 size64, int blocksize, u32 flags) {
 
 void set_timeout(int nbd, int timeout) {
 	if (timeout) {
-#ifdef NBD_SET_TIMEOUT
 		if (ioctl(nbd, NBD_SET_TIMEOUT, (unsigned long)timeout) < 0)
 			err("Ioctl NBD_SET_TIMEOUT failed: %m\n");
 		fprintf(stderr, "timeout=%d\n", timeout);
-#else
-		err("Ioctl NBD_SET_TIMEOUT cannot be called when compiled on a system that does not support it\n");
-#endif
 	}
 }
 
@@ -267,25 +368,43 @@ void finish_sock(int sock, int nbd, int swap) {
 		mlockall(MCL_CURRENT | MCL_FUTURE);
 }
 
+static int
+oom_adjust(const char *file, const char *value)
+{
+	int fd, rc;
+	size_t len;
+
+	fd = open(file, O_WRONLY);
+	if (fd < 0)
+		return -1;
+	len = strlen(value);
+	rc = write(fd, value, len) != (ssize_t) len;
+	close(fd);
+	return rc ? -1 : 0;
+}
+
 void usage(char* errmsg, ...) {
 	if(errmsg) {
 		char tmp[256];
 		va_list ap;
 		va_start(ap, errmsg);
 		snprintf(tmp, 256, "ERROR: %s\n\n", errmsg);
-		vfprintf(stderr, errmsg, ap);
+		vfprintf(stderr, tmp, ap);
 		va_end(ap);
 	} else {
 		fprintf(stderr, "nbd-client version %s\n", PACKAGE_VERSION);
 	}
-	fprintf(stderr, "Usage: nbd-client host port nbd_device [-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S] [-persist|-p] [-nofork|-n] [-name|-N name]\n");
+	fprintf(stderr, "Usage: nbd-client host port nbd_device [-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S] [-persist|-p] [-nofork|-n]\n");
+	fprintf(stderr, "Or   : nbd-client -name|-N name host [port] nbd_device [-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S] [-persist|-p] [-nofork|-n]\n");
 	fprintf(stderr, "Or   : nbd-client -d nbd_device\n");
 	fprintf(stderr, "Or   : nbd-client -c nbd_device\n");
 	fprintf(stderr, "Or   : nbd-client -h|--help\n");
+	fprintf(stderr, "Or   : nbd-client -l|--list host\n");
 	fprintf(stderr, "Default value for blocksize is 1024 (recommended for ethernet)\n");
 	fprintf(stderr, "Allowed values for blocksize are 512,1024,2048,4096\n"); /* will be checked in kernel :) */
 	fprintf(stderr, "Note, that kernel 2.4.2 and older ones do not work correctly with\n");
 	fprintf(stderr, "blocksizes other than 1024 without patches\n");
+	fprintf(stderr, "Default value for port with -N is 10809. Note that port must always be numeric\n");
 }
 
 void disconnect(char* device) {
@@ -293,18 +412,10 @@ void disconnect(char* device) {
 
 	if (nbd < 0)
 		err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
-	printf("Disconnecting: que, ");
-	if (ioctl(nbd, NBD_CLEAR_QUE)< 0)
-		err("Ioctl failed: %m\n");
 	printf("disconnect, ");
-#ifdef NBD_DISCONNECT
 	if (ioctl(nbd, NBD_DISCONNECT)<0)
 		err("Ioctl failed: %m\n");
 	printf("sock, ");
-#else
-	fprintf(stderr, "Can't disconnect: I was not compiled with disconnect support!\n" );
-	exit(1);
-#endif
 	if (ioctl(nbd, NBD_CLEAR_SOCK)<0)
 		err("Ioctl failed: %m\n");
 	printf("done\n");
@@ -326,11 +437,17 @@ int main(int argc, char *argv[]) {
 	int c;
 	int nonspecial=0;
 	char* name=NULL;
+	uint32_t needed_flags=0;
+	uint32_t cflags=0;
+	uint32_t opts=0;
+	sigset_t block, old;
+	struct sigaction sa;
 	struct option long_options[] = {
 		{ "block-size", required_argument, NULL, 'b' },
 		{ "check", required_argument, NULL, 'c' },
 		{ "disconnect", required_argument, NULL, 'd' },
 		{ "help", no_argument, NULL, 'h' },
+		{ "list", no_argument, NULL, 'l' },
 		{ "name", required_argument, NULL, 'N' },
 		{ "nofork", no_argument, NULL, 'n' },
 		{ "persist", no_argument, NULL, 'p' },
@@ -342,7 +459,7 @@ int main(int argc, char *argv[]) {
 
 	logging();
 
-	while((c=getopt_long_only(argc, argv, "-b:c:d:hnN:pSst:", long_options, NULL))>=0) {
+	while((c=getopt_long_only(argc, argv, "-b:c:d:hlnN:pSst:", long_options, NULL))>=0) {
 		switch(c) {
 		case 1:
 			// non-option argument
@@ -374,10 +491,6 @@ int main(int argc, char *argv[]) {
 						nonspecial++;
 					} else {
 						port = optarg;
-						if(name) {
-							usage("port and name specified at the same time. This is not supported.");
-							exit(EXIT_FAILURE);
-						}
 					}
 					break;
 				case 2:
@@ -401,16 +514,22 @@ int main(int argc, char *argv[]) {
 		case 'h':
 			usage(NULL);
 			exit(EXIT_SUCCESS);
+		case 'l':
+			needed_flags |= NBD_FLAG_FIXED_NEWSTYLE;
+			cflags |= NBD_FLAG_C_FIXED_NEWSTYLE;
+			opts |= NBDC_DO_LIST;
+			name="";
+			nbddev="";
+			port = NBD_DEFAULT_PORT;
+			break;
 		case 'n':
 			nofork=1;
 			break;
 		case 'N':
 			name=optarg;
-			if(port) {
-				usage("port and name specified at the same time. This is not supported.");
-				exit(EXIT_FAILURE);
+			if(!port) {
+				port = NBD_DEFAULT_PORT;
 			}
-			port = NBD_DEFAULT_PORT;
 			break;
 		case 'p':
 			cont=1;
@@ -436,17 +555,26 @@ int main(int argc, char *argv[]) {
 		exit(EXIT_FAILURE);
 	}
 
+	sock = opennet(hostname, port, sdp);
+	if (sock < 0)
+		exit(EXIT_FAILURE);
+
+	negotiate(sock, &size64, &flags, name, needed_flags, cflags, opts);
+
 	nbd = open(nbddev, O_RDWR);
 	if (nbd < 0)
 	  err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
-	++argv; --argc; /* skip device */
 
-	sock = opennet(hostname, port, sdp);
-
-	negotiate(sock, &size64, &flags, name);
 	setsizes(nbd, size64, blocksize, flags);
 	set_timeout(nbd, timeout);
 	finish_sock(sock, nbd, swap);
+	if (swap) {
+		/* try linux >= 2.6.36 interface first */
+		if (oom_adjust("/proc/self/oom_score_adj", "-1000")) {
+			/* fall back to linux <= 2.6.35 interface */
+			oom_adjust("/proc/self/oom_adj", "-17");
+		}
+	}
 
 	/* Go daemon */
 	
@@ -455,10 +583,21 @@ int main(int argc, char *argv[]) {
 		if (daemon(0,0) < 0)
 			err("Cannot detach from terminal");
 	}
+
+	memset(&sa, 0, sizeof(sa));
+	sa.sa_handler = SIG_IGN;
+	sigaction(SIGCHLD, &sa, NULL);
 #endif
 	do {
 #ifndef NOFORK
-		if (fork()) {
+
+		sigfillset(&block);
+		sigdelset(&block, SIGKILL);
+		sigdelset(&block, SIGTERM);
+		sigdelset(&block, SIGPIPE);
+		sigprocmask(SIG_SETMASK, &block, &old);
+
+		if (!fork()) {
 			/* Due to a race, the kernel NBD driver cannot
 			 * call for a reread of the partition table
 			 * in the handling of the NBD_DO_IT ioctl().
@@ -479,8 +618,9 @@ int main(int argc, char *argv[]) {
 #endif
 
 		if (ioctl(nbd, NBD_DO_IT) < 0) {
-			fprintf(stderr, "Kernel call returned: %m");
-			if(errno==EBADR) {
+		        int error = errno;
+			fprintf(stderr, "nbd,%d: Kernel call returned: %d", getpid(), error);
+			if(error==EBADR) {
 				/* The user probably did 'nbd-client -d' on us.
 				 * quit */
 				cont=0;
@@ -489,11 +629,18 @@ int main(int argc, char *argv[]) {
 					u64 new_size;
 					u32 new_flags;
 
-					fprintf(stderr, " Reconnecting\n");
 					close(sock); close(nbd);
-					sock = opennet(hostname, port, sdp);
+					for (;;) {
+						fprintf(stderr, " Reconnecting\n");
+						sock = opennet(hostname, port, sdp);
+						if (sock >= 0)
+							break;
+						sleep (1);
+					}
 					nbd = open(nbddev, O_RDWR);
-					negotiate(sock, &new_size, &new_flags, name);
+					if (nbd < 0)
+						err("Cannot open NBD: %m");
+					negotiate(sock, &new_size, &new_flags, name, needed_flags, cflags, opts);
 					if (size64 != new_size) {
 						err("Size of the device changed. Bye");
 					}
@@ -512,8 +659,6 @@ int main(int argc, char *argv[]) {
 			cont=0;
 		}
 	} while(cont);
-	printf("Closing: que, ");
-	ioctl(nbd, NBD_CLEAR_QUE);
 	printf("sock, ");
 	ioctl(nbd, NBD_CLEAR_SOCK);
 	printf("done\n");
